# 动态代理

## 前言

先说普遍的知识点，动态代理是实现AOP的方法，分为两种，JDK原生和CGLIB方法，两者**主要区别**在于：

- JDK动态代理是实现了被代理对象的接口，Cglib是继承了被代理对象，对于Cglib来说，目标对象有没有实现接口都可以代理，但不能代理final方法，因为不能继承。JDK和Cglib都是在运行期生成字节码，
- JDK是直接写Class字节码，Cglib使用ASM框架写Class字节码，Cglib代理实现更复杂，生成代理类比JDK效率低。
- JDK调用代理方法，是通过反射机制生成一个实现代理接口的匿名类，在调用具体方法前调用InvokeHandler来处理。而Cglib是通过FastClass机制直接调用方法，Cglib执行效率更高。

在Spring中，当需要使用AOP的类具有接口时，它会以JDK动态代理运行，否则以CGLIB运行。

另外，还有一些细节差别：

- JDK动态代理，类中内部方法间的调用，被调用的方法不会被代理。
- CGLIB动态代理，类中内部方法间的调用，被调用的方法会被代理。

**我们结合具体实例，来进行具体细节讲解。**

首先新建一个用于代理的接口（后续委托类和代理类的公共接口）。

```java
public interface UserService {
    public void printUser();
    public void vaildUser();
}
```

以及实现类（后续的委托类）。

```java
public class UserServiceImpl implements UserService {
    @Override
    public void printUser(){
        vaildUser();//类内方法调用
        System.out.println("I'm Zingbug");
    }
    @Override
    public void vaildUser() {
        System.out.println("I'm real");
    }
}
```

## 源码分析

创建代理类，需要实现InvocationHandler接口，并重写invoke方法。

```java
public class ProxyBean implements InvocationHandler {
    private Object target=null;
    public static Object getProxyBean(Object target){
        ProxyBean proxyBean=new ProxyBean();
        proxyBean.target=target;
        Object proxy= Proxy.newProxyInstance(target.getClass().getClassLoader(),target.getClass().getInterfaces(),proxyBean);//生成代理对象  注意这行！！！
        return proxy;
    }
    @Override
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        System.out.println("before JDK ...");
        method.invoke(target,args);//反射调用  注意这行！！！
        System.out.println("after JDK ...");
        return null;
    }
}
```

生成代理类的关键在于方法：

```java
Proxy.newProxyInstance(ClassLoader loader,Class<?>[] interfaces,InvocationHandler h)
```

在这个方法内，主要流程为：

**1、根据类加载器classLoader和接口数组来生成代理类的class对象。**

```java
final Class<?>[] intfs = interfaces.clone();
Class<?> cl = getProxyClass0(loader, intfs);//二进制的class文件
```

**2、获取代理类中参数为InvocationHandler的构造方法。**

```java
//private static final Class<?>[] constructorParams = { InvocationHandler.class };
final Constructor<?> cons = cl.getConstructor(constructorParams);
```

**3、判断代理类cl代理类Class的修饰符是不是pulic类型，如果是，则传入自定义的InvocationHandler，并返回指定构造器cons的实例，即代理对象的实例。如果不是，则将cons构造方法设置为可访问。**

```java
if (!Modifier.isPublic(cl.getModifiers())) {
    AccessController.doPrivileged(new PrivilegedAction<Void>() {
        public Void run() {
            cons.setAccessible(true);
            return null;
        }
    });
}
return cons.newInstance(new Object[]{h});//传入自定义的InvocationHandler
```

另外，还需要看一下通过类加载器和接口数组来生成代理类字节码的方法getProxyClass0（**使用了缓存技术，只需要生成一次**）。

```java
private static final WeakCache<ClassLoader, Class<?>[], Class<?>>
    proxyClassCache = new WeakCache<>(new KeyFactory(), new ProxyClassFactory());

private static Class<?> getProxyClass0(ClassLoader loader,
                                       Class<?>... interfaces) {
    if (interfaces.length > 65535) {
        throw new IllegalArgumentException("interface limit exceeded");
    }
    // If the proxy class defined by the given loader implementing
    // the given interfaces exists, this will simply return the cached copy;
    // otherwise, it will create the proxy class via the ProxyClassFactory
    return proxyClassCache.get(loader, interfaces);
}
```

可知，proxyClassCache是WeakCache类型，如果缓存中包含接口数组和classloader的class类已经存在，就返回缓存副本，否则通过ProxyClassFactory创建一个对应的class对象。

到此为止，讲完了代理类ProxyBean的生成流程，那既然已经实现了InvocationHandler接口，就需要重写**invoke方法**。

在invoke(Object proxy, Method method, Object[] args)方法中利用**反射调用**实际的方法。注意，是委托类的方法，而不是代理类的方法。这也是前面提到的JDK动态代理中，内部被调用的方法不会被代理，因为调用的printUser方法是由委托类target执行，里面调用的vaildUser方法不会被代理。

```java
method.invoke(target,args);//反射调用
```

target是委托类，也就是UserServiceImpl。而传入参数中的proxy是代理的字节码类。

代理类完事后，下面进行测试。

```java
public static void testJdk() {
    UserService userService = (UserService) ProxyBean.getProxyBean(new UserServiceImpl());
    userService.printUser();

    //将代理类字节码保存下来
    byte[] classFile= ProxyGenerator.generateProxyClass("$Proxy0",UserServiceImpl.class.getInterfaces());
    String path="自定义保存路径";
    try {
        FileOutputStream fos=new FileOutputStream(path);
        fos.write(classFile);
        fos.flush();
        System.out.println("写入成功");
    }
    catch (Exception e){
        System.out.println("写入错误");
    }
}

运行结果：
before JDK ...
I'm real
I'm Zingbug
after JDK ...
```

我们可以看出内部被调用的方法vaildUser()没有被代理，符合之前的结论和原因分析。另外，我们将生成的字节码类文件保留下来，编译出来分析一波。

```java
public final class $Proxy0 extends Proxy implements UserService {//继承自Proxy，实现了UserService接口  注意这行！！！
    private static Method m1;
    private static Method m4;
    private static Method m2;
    private static Method m0;
    private static Method m3;
    public $Proxy0(InvocationHandler var1) throws  {
        super(var1);
    }
    public final boolean equals(Object var1) throws  {
        try {
            return (Boolean)super.h.invoke(this, m1, new Object[]{var1});
        } catch (RuntimeException | Error var3) {
            throw var3;
        } catch (Throwable var4) {
            throw new UndeclaredThrowableException(var4);
        }
    }
    public final void printUser() throws  {//注意这行！！！
        try {
            super.h.invoke(this, m4, (Object[])null);//注意这行！！！
        } catch (RuntimeException | Error var2) {
            throw var2;
        } catch (Throwable var3) {
            throw new UndeclaredThrowableException(var3);
        }
    }
    public final void vaildUser() throws  {
        try {
            super.h.invoke(this, m3, (Object[])null);
        } catch (RuntimeException | Error var2) {
            throw var2;
        } catch (Throwable var3) {
            throw new UndeclaredThrowableException(var3);
        }
    }
    //省略部分代理的Object方法源码
    static {
        try {
            m1 = Class.forName("java.lang.Object").getMethod("equals", Class.forName("java.lang.Object"));
            m4 = Class.forName("com.example.demo.service.UserService").getMethod("printUser");//注意这行！！！
            m2 = Class.forName("java.lang.Object").getMethod("toString");
            m0 = Class.forName("java.lang.Object").getMethod("hashCode");
            m3 = Class.forName("com.example.demo.service.UserService").getMethod("vaildUser");//注意这行！！！
        } catch (NoSuchMethodException var2) {
            throw new NoSuchMethodError(var2.getMessage());
        } catch (ClassNotFoundException var3) {
            throw new NoClassDefFoundError(var3.getMessage());
        }
    }
}
```

从这个字节码中，可以发现代理类继承自Proxy，实现了UserService方法，其构造函数就是传递了InvocationHandler参数，同时实现了Object类的equals ，hashcode，toString方法，以及UserService接口自己定义的printUser和vaildUser方法。并且，所有接口方法的实现都委托给InvocationHandler的invoke方法了，这也就是实现代理模式的地方了。

**下面来看CGLIB**，其中被代理的类还是UserServiceImpl类。

**1、定义一个拦截器。在调用目标方法时，CGLib会回调MethodInterceptor接口方法拦截，来实现自己的代理逻辑，类似于JDK中的InvocationHandler接口。**

```java
public class UserCglib implements MethodInterceptor {
    @Override
    public Object intercept(Object o, Method method, Object[] objects, MethodProxy methodProxy) throws Throwable {
        System.out.println("before CGLIB ...");
        Object value=methodProxy.invokeSuper(o,objects);//不要用methodProxy.invoke()  注意这行！！！
        System.out.println("after CGLIB ...");
        return value;
    }
}
```

看一下intercept方法入参：

- o：cglib生成的代理对象，
- method：被代理对象方法，
- objects：方法入参，
- methodProxy:代理方法。

**2、生成动态代理类，并在代理类上进行方法测试。**

```java
public static void testCglib()
{
    System.setProperty(DebuggingClassWriter.DEBUG_LOCATION_PROPERTY,"自定义路径");//将代理类的字节码保存下来
    Enhancer enhancer=new Enhancer();
    enhancer.setSuperclass(UserServiceImpl.class);
    enhancer.setCallback(new UserCglib());
    UserService userService=(UserService)enhancer.create();
    userService.printUser();
}

运行结果：
before CGLIB ...
before CGLIB ...
I'm real
after CGLIB ...
I'm Zingbug
after CGLIB ...
```

这里Enhancer是CGLib的一个字节码增强器，它可以方便的对你想要处理的类进行扩展。首先将被代理类UserServiceImpl设置成父类，然后设置拦截器UserCglib，最后执行enhancer.create()动态生成一个代理类，并从Object强制转型成父类型UserServiceImpl。

可以看出，与JDK原生方法不同的是，这次内部被调用的方法vaildUser()也被代理了。

执行testCglib()测试可以得到CGLib生成的class文件，一共有三个class文件，反编译以供学习。

![Y43vMF.png](https://s1.ax1x.com/2020/05/19/Y43vMF.png)

其中UserServiceImpl$$EnhancerByCGLIB$$dd4bde41.class就是CGLIB生成的代理类，它继承了UserServiceImpl类。无用代码不再展示。

```java
public class UserServiceImpl$$EnhancerByCGLIB$$dd4bde41 extends UserServiceImpl implements Factory {
    private boolean CGLIB$BOUND;
    public static Object CGLIB$FACTORY_DATA;
    private static final ThreadLocal CGLIB$THREAD_CALLBACKS;
    private static final Callback[] CGLIB$STATIC_CALLBACKS;
    private MethodInterceptor CGLIB$CALLBACK_0;//拦截器    注意这行！！！
    private static Object CGLIB$CALLBACK_FILTER;
    private static final Method CGLIB$printUser$0$Method;//被代理方法printUser()   注意这行！！！
    private static final MethodProxy CGLIB$printUser$0$Proxy;
    private static final Object[] CGLIB$emptyArgs;
    private static final Method CGLIB$vaildUser$1$Method;//被代理方法vaildUser()    注意这行！！！
    private static final MethodProxy CGLIB$vaildUser$1$Proxy;
    private static final Method CGLIB$equals$2$Method;
    private static final MethodProxy CGLIB$equals$2$Proxy;
    private static final Method CGLIB$toString$3$Method;
    private static final MethodProxy CGLIB$toString$3$Proxy;
    private static final Method CGLIB$hashCode$4$Method;
    private static final MethodProxy CGLIB$hashCode$4$Proxy;
    private static final Method CGLIB$clone$5$Method;
    private static final MethodProxy CGLIB$clone$5$Proxy;
    static void CGLIB$STATICHOOK1() {
        CGLIB$THREAD_CALLBACKS = new ThreadLocal();
        CGLIB$emptyArgs = new Object[0];
        Class var0 = Class.forName("com.example.demo.service.UserServiceImpl$$EnhancerByCGLIB$$dd4bde41");//代理类      注意这行！！！
        Class var1;//委托类UserServiceImpl       注意这行！！！
        Method[] var10000 = ReflectUtils.findMethods(new String[]{"equals", "(Ljava/lang/Object;)Z", "toString", "()Ljava/lang/String;", "hashCode", "()I", "clone", "()Ljava/lang/Object;"}, (var1 = Class.forName("java.lang.Object")).getDeclaredMethods());
        CGLIB$equals$2$Method = var10000[0];
        CGLIB$equals$2$Proxy = MethodProxy.create(var1, var0, "(Ljava/lang/Object;)Z", "equals", "CGLIB$equals$2");
        CGLIB$toString$3$Method = var10000[1];
        CGLIB$toString$3$Proxy = MethodProxy.create(var1, var0, "()Ljava/lang/String;", "toString", "CGLIB$toString$3");
        CGLIB$hashCode$4$Method = var10000[2];
        CGLIB$hashCode$4$Proxy = MethodProxy.create(var1, var0, "()I", "hashCode", "CGLIB$hashCode$4");
        CGLIB$clone$5$Method = var10000[3];
        CGLIB$clone$5$Proxy = MethodProxy.create(var1, var0, "()Ljava/lang/Object;", "clone", "CGLIB$clone$5");
        var10000 = ReflectUtils.findMethods(new String[]{"printUser", "()V", "vaildUser", "()V"}, (var1 = Class.forName("com.example.demo.service.UserServiceImpl")).getDeclaredMethods()); //注意这行！！！
        CGLIB$printUser$0$Method = var10000[0];   //注意这行！！
        CGLIB$printUser$0$Proxy = MethodProxy.create(var1, var0, "()V", "printUser", "CGLIB$printUser$0");   //注意这行！！
        CGLIB$vaildUser$1$Method = var10000[1];  //注意这行！！
        CGLIB$vaildUser$1$Proxy = MethodProxy.create(var1, var0, "()V", "vaildUser", "CGLIB$vaildUser$1");    }    //注意这行！！
    //代理方法（methodProxy.invokeSuper会调用）
    final void CGLIB$printUser$0() {        super.printUser();    }
    //代理方法（methodProxy.invoke会调用，这就是为什么在拦截器中调用methodProxy.invoke会死循环，一直在调用拦截器）
    public final void printUser() {
        MethodInterceptor var10000 = this.CGLIB$CALLBACK_0;
        if (var10000 == null) {
            CGLIB$BIND_CALLBACKS(this);
            var10000 = this.CGLIB$CALLBACK_0;
        }
        if (var10000 != null) {
            var10000.intercept(this, CGLIB$printUser$0$Method, CGLIB$emptyArgs, CGLIB$printUser$0$Proxy);
        } else {
            super.printUser();
        }
    }
    final void CGLIB$vaildUser$1() {
        super.vaildUser();
    }
    public final void vaildUser() {
        MethodInterceptor var10000 = this.CGLIB$CALLBACK_0;
        if (var10000 == null) {
            CGLIB$BIND_CALLBACKS(this);
            var10000 = this.CGLIB$CALLBACK_0;
        }
        if (var10000 != null) {
            var10000.intercept(this, CGLIB$vaildUser$1$Method, CGLIB$emptyArgs, CGLIB$vaildUser$1$Proxy);
        } else {
            super.vaildUser();
        }
    }
    .......
}
```

通过代理类的源码可以看到，代理类会获得所有在父类继承来的方法（包括所有的Object方法），并且会有MethodProxy与之对应，比如 CGLIB$printUser$0$Method、CGLIB$printUser$0$Proxy等。

对于方法调用过程：代理对象调用this.printUser()方法 -> 调用拦截器 -> methodProxy.invokeSuper函数 -> CGLIB$printUser$0()方法 -> 委托类对象的printUser()方法。

拦截器MethodInterceptor中就是由MethodProxy的invokeSuper方法调用代理方法的，MethodProxy非常关键。具体看一下这个代码。

1、**首先，创建MethodProxy。**

```java
public static MethodProxy create(Class c1, Class c2, String desc, String name1, String name2) {
    MethodProxy proxy = new MethodProxy();
    proxy.sig1 = new Signature(name1, desc);
    proxy.sig2 = new Signature(name2, desc);
    proxy.createInfo = new MethodProxy.CreateInfo(c1, c2);
    return proxy;
}
```

其中，c1是委托对象Class，c2是代理对象Class，desc是入参类型，name1是委托方法名，name2是代理方法名。

2、**对于invokeSuper调用。**

```java
public Object invokeSuper(Object obj, Object[] args) throws Throwable {
    try {
        this.init();//初始化，生成FastClass，并放入缓存。  注意这行！！！
        MethodProxy.FastClassInfo fci = this.fastClassInfo;
        return fci.f2.invoke(fci.i2, obj, args);//直接调用方法  注意这行！！！
    } catch (InvocationTargetException var4) {
        throw var4.getTargetException();
    }
}

private static class FastClassInfo {
    FastClass f1;
    FastClass f2;
    int i1;
    int i2;
    private FastClassInfo() {
    }
}
```

代码调用过程就是获取到代理类对应的FastClass，并执行了代理方法。还记得之前生成三个class文件吗？

UserServiceImpl$$EnhancerByCGLIB$$dd4bde41$$FastClassByCGLIB$$6ad9babb.class就是代理类的

FastClass，UserServiceImpl$$FastClassByCGLIB$$ff73797.class就是委托类的FastClass。

3、**下面这就引入了FastClass机制。**

Cglib动态代理执行代理方法效率之所以比JDK的高是因为Cglib采用了FastClass机制，它的原理简单来说就是：为代理类和被代理类各生成一个Class，这个Class会为代理类或被代理类的方法分配一个index(int类型)。

这个index当做一个入参，FastClass就可以直接定位要调用的方法直接进行调用，这样**省去了反射调用**，所以调用效率比JDK动态代理通过反射调用高。下面我们反编译一个FastClass看看：

```java
//根据方法签名获取index
public int getIndex(Signature var1) {
    String var10000 = var1.toString();
    switch(var10000.hashCode()) {
    case -1544526380:
        if (var10000.equals("vaildUser()V")) {
            return 1;
        }
        break;
    case -7204771:
        if (var10000.equals("printUser()V")) {
            return 0;
        }
        break;
        //省略其他Object类方法代码....
    }
    return -1;
}

//根据index直接定位执行方法
public Object invoke(int var1, Object var2, Object[] var3) throws InvocationTargetException {
    UserServiceImpl var10000 = (UserServiceImpl)var2;
    int var10001 = var1;
    try {
        switch(var10001) {
        case 0:
            var10000.printUser();//注意这行！！！
            return null;
        case 1:
            var10000.vaildUser();//注意这行！！！
            return null;
        case 2:
            return new Boolean(var10000.equals(var3[0]));
        case 3:
            return var10000.toString();
        case 4:
            return new Integer(var10000.hashCode());
        }
    } catch (Throwable var4) {
        throw new InvocationTargetException(var4);
    }
    throw new IllegalArgumentException("Cannot find matching method/constructor");
}
```

其中，FastClass并不是跟代理类一块生成的，而是在第一次执行MethodProxy invoke/invokeSuper时生成的并放在了缓存中。

```java
//MethodProxy invoke/invokeSuper都调用了init()
private void init() {
    if (this.fastClassInfo == null) {
        synchronized(this.initLock) {
            if (this.fastClassInfo == null) {
                MethodProxy.CreateInfo ci = this.createInfo;
                MethodProxy.FastClassInfo fci = new MethodProxy.FastClassInfo();
                fci.f1 = helper(ci, ci.c1);//如果缓存中就取出，没有就生成新的FastClass 注意这行！！！
                fci.f2 = helper(ci, ci.c2);
                fci.i1 = fci.f1.getIndex(this.sig1);//获取方法的index 注意这行！！！
                fci.i2 = fci.f2.getIndex(this.sig2);
                this.fastClassInfo = fci;
                this.createInfo = null;
            }
        }
    }
}
```

至此，CGLIB动态代理原理就差不多搞清楚了，还是深究源码才能弄懂。

## 参考链接

[jdk和cglib原理区别](https://blog.csdn.net/weixin_39158271/article/details/78074442)

[JDK动态代理实现原理——JDK1.8](https://blog.csdn.net/huangwei18351/article/details/82460589)

[CGLIB动态代理实现方式](https://www.cnblogs.com/monkey0307/p/8328821.html)

弥有，2019/7/25
[EOF]
