# 单例模式

本文将一步步介绍单例模式。

## 1、懒汉模式

```java
public class Singleton {
    private static final Singleton instance = new Singleton();
    //私有的默认构造函数
    private Singleton() {
    }
    //静态工厂方法
    public static Singleton getInstance() {
        return instance;
    }
}
```

这是最简单的，缺点很明显，第一次加载类的时候会连带着创建Singleton实例，这样的结果与我们所期望的不同，因为创建实例的时候可能并不是我们需要这个实例的时候。同时如果这个Singleton实例的创建非常消耗系统资源，而应用始终都没有使用Singleton实例，那么创建Singleton消耗的系统资源就被白白浪费了。但是也有优点，这个是线程安全的。

为了避免这种情况，我们通常使用**惰性加载** 的机制，也就是在使用的时候才去创建。但简单的惰性加载会有线程安全问题，一般会加入锁来保证每次只有唯一线程获取实例。

## 2、饿汉模式

```java
public class Singleton {
    private static Singleton instance = null;
    //私有的默认构造函数
    private Singleton() {
    }
    //静态工厂方法
    public synchronized static Singleton getInstance() {
        if (instance == null) {
            instance = new Singleton();
        }
        return instance;
    }
}
```

这种方法解决了多线程并发安全问题，但是它却很影响性能，每次调用getInstance()方法的时候都必须获得Singleton类的锁（静态方法，synchronized锁针对的是类）。而实际上，当单例实例被创建后，其后的请求没有必要再使用互斥机制了。为了解决这个问题，提出了double-checked locking的解决方案。

## 3、双重检查锁

```java
public static Singleton getInstance() {
    if (instance == null) {
        synchronized (instance) {
            if (instance == null) {
                instance = new Singleton();
            }
        }
    }
    return instance;
}
```

当一个线程发出请求后，会先检查instance是否为null，如果不是则直接返回其内容，这样避免了进入synchronized块所需要花费的资源。这个方案似乎解决了面临的问题，但从JVM角度来说，这些代码依旧可能发生错误。

对于JVM而言，它执行的是一个个Java指令。在Java指令中创建对象和赋值操作是分开进行的，也就是说instance = new Singleton();语句是分两步执行的。但是JVM并不保证这两个操作的先后顺序，也就是说有可能JVM会为新的Singleton实例分配空间，然后直接赋值给instance成员，然后再去初始化这个Singleton实例。(即先赋值指向了内存地址，再初始化)这样就使出错成为了可能，我们以A、B两个线程为例：

1. A、B线程同时进入了第一个if判断。
2. A首先进入synchronized块，由于instance为null，所以它执行instance = new Singleton();
3. 由于JVM内部的优化机制，JVM先画出了一些分配给Singleton实例的空白内存，并赋值给instance成员（注意此时JVM没有开始初始化这个实例），然后A离开了synchronized块。
4. B进入synchronized块，由于instance此时不是null，因此它马上离开了synchronized块并将结果返回给调用该方法的程序。
5. 此时B线程打算使用Singleton实例，却发现它没有被初始化，于是错误发生了。

## 4、内部类

```java
public class Singleton {
    //私有的默认构造函数
    private Singleton() {
    }
    //内部类
    private static class SingletonContainer {
        private static Singleton instance = new Singleton();
    }
    //静态工厂方法
    public static Singleton getInstance() {
        return SingletonContainer.instance;
    }
}
```

JVM内部的机制能够保证当一个类被加载的时候，这个类的加载过程是线程互斥的。这样当我们第一次调用getInstance的时候，JVM能够帮我们保证instance只被创建一次，并且会保证把赋值给instance的内存初始化完毕，避免了双重检查方案的问题。此外该方法也只会在第一次调用的时候使用互斥机制，这样就解决了懒汉模式的低效问题。最后instance是在第一次加载SingletonContainer类时被创建的，而SingletonContainer类则在调用getInstance方法的时候才会被加载，因此也实现了惰性加载。

弥有，2019年7月
[EOF]
