# 循环依赖

## 问题概述

### 什么是循环依赖

Bean A 依赖 B，Bean B 依赖 A这种情况下出现循环依赖。

Bean A → Bean B → Bean A

更复杂的间接依赖造成的循环依赖如下。

Bean A → Bean B → Bean C → Bean D → Bean E → Bean A

### 循环依赖会产生什么结果

当Spring正在加载所有Bean时，Spring尝试以能正常创建Bean的顺序去创建Bean。

例如，有如下依赖:

Bean A → Bean B → Bean C

Spring先创建beanC，接着创建bean B（将C注入B中)，最后创建bean A(将B注入A中)。

但当存在循环依赖时，Spring将无法决定先创建哪个bean。这种情况下，Spring将产生异常BeanCurrentlyInCreationException。

当使用**构造器注入**时经常会发生循环依赖问题。如果使用其它类型的注入方式能够避免这种问题。

下面举个例子。首先定义两个相互通过构造器注入依赖的bean。

```java
@Component
public class CircularDependencyA {
    private CircularDependencyB circB;
    @Autowired
    public CircularDependencyA(CircularDependencyB circB) {//构造器注入依赖
        this.circB = circB;
    }
}
```

```java
@Component
public class CircularDependencyB {
    private CircularDependencyA circA;
    @Autowired
    public CircularDependencyB(CircularDependencyA circA) {
        this.circA = circA;
    }
}
```

写一个配置类，保证能够扫描到上面创建的bean。

```java
@Configuration
@ComponentScan(basePackages = { "com.baeldung.circulardependency" })
public class TestConfig {
}
```

最后写一个测试案例

```java
@RunWith(SpringJUnit4ClassRunner.class)
@ContextConfiguration(classes = { TestConfig.class })
public class CircularDependencyTest {
    @Test
    public void givenCircularDependency_whenConstructorInjection_thenItFails() {
        // Empty test; we just want the context to load
    }
}
```

运行方法CircularDependencyTest将会产生异常：

```java
The dependencies of some of the beans in the application context form a cycle:
┌────┐
|  circa defined in file [CircularDependencyA.class]
↑    ↓
|  circb defined in file [CircularDependencyB.class]
└────┘
```

## 解决方法

处理这种问题有几种常见的方式。

### 1、重新设计

重新设计结构，消除循环依赖。

### 2、使用懒注解@Lazy

最简单的消除循环依赖的方式是通过延迟加载。具体代码如下。

```java
@Component
public class CircularDependencyA {
    private CircularDependencyB circB;
    @Autowired
    public CircularDependencyA(@Lazy CircularDependencyB circB) {//先被初始化
        this.circB = circB;
    }
}
```

CircularDependencyB是被延迟加载了，它在注入CircularDependencyA中时，并没有完全初始化，而仅是通过一个代理（本文案例中是CGLIB动态代理）将它注入到CircularDependencyA。而CircularDependencyB只有在第一次需要时才会完全创建。

总结为：对于单实例bean，默认是在容器启动的时候创建对象。但当使用懒加载时，将对象的创建推迟到第一次获取的时候。

### 3、直接使用Autowired单独注入

直接使用@Autowired注入依赖，不要使用构造器的方式注入，如下。

```Java
@Component
public class CircularDependencyA {
    @Autowired
    private CircularDependencyB circB;//直接注入
}
```

```java
@Component
public class CircularDependencyB {
    @Autowired
    private CircularDependencyA circA;//直接注入
}
```

### 4、使用Setter注入

如下：

```java
@Component
public class CircularDependencyA {
    private CircularDependencyB circB;
    @Autowired
    public void setCircB(CircularDependencyB circB) {//Setter注入
        this.circB = circB;
    }
}
```

```java
@Component
public class CircularDependencyB {
    private CircularDependencyA circA;
    @Autowired
    public void setCircA(CircularDependencyA circA) {
        this.circA = circA;
    }
}
```

另外，还有一些其他方法，比如使用@PostConstruct等等。

## 总结

最后，当遇到循环依赖时。首先考虑是否能够通过重新设计依赖来避免循环依赖。如果确实需要循环依赖，那么可以通过前文提到的方式来处理。优先建议使用setter注入来解决。

## 参考链接

[Circular Dependencies in Spring](https://www.baeldung.com/circular-dependencies-in-spring)

弥有，2019/7/26
[EOF]
